import select
import socket
import os
import struct
import errno
import time


class DuctworksException(Exception):
    pass


class CommunicationFaultException(DuctworksException):
    pass


class AlreadyConnectedException(DuctworksException):
    pass


class NotConnectedException(DuctworksException):
    pass


class RemoteSocketClosed(DuctworksException):
    pass


def unix_domain_socket_constructor(linger_time=3):
    """
    Create a new UDS streaming socket with reasonable socket options set.

    :param linger_time: The linger time after closing the socket to allow buffers to flush. Default: 3
    :type linger_time: int
    :return: A new Unix Domain socket.
    :rtype: socket.socket
    """
    new_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    new_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    new_socket.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, struct.pack('ii', 1, linger_time))
    return new_socket


def tcp_socket_constructor(linger_time=10, tcp_no_delay=1):
    """
    Create a new TCP socket with reasonable socket options set.

    :param linger_time: The linger time after closing the socket to allow buffers to flush. Default: 10
    :type linger_time: int
    :param tcp_no_delay: 1 -> disable Nagle's algorithm on the created socket, 0 -> enable Nagle's algorithm.
        Default: 1
    :type tcp_no_delay: int
    :return: A new TCP socket.
    :rtype socket.socket
    """
    new_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    new_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    new_socket.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, struct.pack('ii', 1, linger_time))
    # Because everything is sent in a single send call, and we're sending "messages", not really "streams",
    # we turn off Nagle's algorithm to make performance a little better. This might we worth thinking about more though.
    new_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, tcp_no_delay)
    return new_socket


def unix_domain_socket_listener_destructor(listener_socket, shutdown_mode=socket.SHUT_RDWR):
    """
    Close and clean up a Unix Domain listener socket.

    :param listener_socket: The socket to close down.
    :type listener_socket: socket.socket
    :param shutdown_mode: The parameters to be passed to the socket's shutdown function.
    :type shutdown_mode: int
    :return: None
    """
    socket_address = listener_socket.getsockname()
    listener_socket.shutdown(shutdown_mode)
    listener_socket.close()
    try:
        os.unlink(socket_address)
    except OSError:
        pass


def tcp_socket_listener_destructor(listener_socket, shutdown_mode=socket.SHUT_RDWR):
    """
    Close and clean up a TCP listener socket.

    :param listener_socket: The socket to close down.
    :type listener_socket: socket.socket
    :param shutdown_mode: The parameters to be passed to the socket's shutdown function.
    :type shutdown_mode: int
    :return: None
    """
    listener_socket.shutdown(shutdown_mode)
    listener_socket.close()


def client_socket_destructor(client_socket, shutdown_mode=socket.SHUT_RDWR):
    """
    Close and clean up a client socket.

    :param client_socket: The socket to close down.
    :type client_socket: socket.socket
    :param shutdown_mode: The parameters to be passed to the socket's shutdown function.
    :type shutdown_mode: int
    :return: None
    """
    client_socket.shutdown(shutdown_mode)
    client_socket.close()


class RawDuctParent(object):
    """
    The RawDuctParent is a thin wrapper over top of a "server" socket, that uses the socket in an "anonymous" way.
    Once the first connection is seen, the listener socket is closed, and the "parent" and "child" act like an
    anonymous socket pair (with a few extra niceties, like a poll method).
    """

    DEFAULT_TIMEOUT = 30

    def __init__(self, bind_address, server_listener_socket_constructor=unix_domain_socket_constructor,
                 server_listener_socket_destructor=unix_domain_socket_listener_destructor,
                 server_connection_socket_destructor=client_socket_destructor, timeout=DEFAULT_TIMEOUT):
        self.server_listener_socket_constructor = server_listener_socket_constructor
        self.server_listener_socket_destructor = server_listener_socket_destructor
        self.server_connection_socket_destructor = server_connection_socket_destructor
        self.bind_address = bind_address
        self.listener_address = None
        self.listener_socket = None
        self.conn_socket = None
        self.socket_timeout = timeout

    def bind(self, listen_queue_depth=1):
        """
        Create and bind the listener socket, if this hasn't been done already.

        This will raise AlreadyConnectedException if the steady-state connection socket has already been created.

        :param listen_queue_depth: The queue depth for the listener socket. Default: 1
        :return: None
        """
        if self.conn_socket is not None:
            raise AlreadyConnectedException("Already connected to other end!")
        if self.listener_socket is None:
            self.listener_socket = self.server_listener_socket_constructor()
            self.listener_socket.settimeout(self.socket_timeout)
            self.listener_socket.bind(self.bind_address)
            self.listener_address = self.listener_socket.getsockname()
            self.listener_socket.listen(listen_queue_depth)

    def listen(self, timeout=60):
        """
        Listen for incoming connections from an incoming child duct.
        This will bind if it hasn't been done already.

        This will raise AlreadyConnectedException if the steady-state connection socket has already been created.

        :param timeout: Amount of time to wait for the other end to connect before giving up.
        :param timeout: float | int
        :return: True if a connection was received and connected, False otherwise.
        :rtype: bool
        """
        if self.conn_socket is not None:
            raise AlreadyConnectedException("Already connected to other end!")
        if self.listener_socket is None:
            self.bind()
        listener_fd = self.listener_socket.fileno()
        has_conn, _, is_faulted = map(bool, select.select([listener_fd], [], [listener_fd], timeout))
        if is_faulted:
            self.server_listener_socket_destructor(self.listener_socket)
            raise CommunicationFaultException("Bind socket faulted!")
        elif has_conn:
            self.conn_socket, _ = self.listener_socket.accept()
            self.conn_socket.settimeout(self.socket_timeout)
            self.server_listener_socket_destructor(self.listener_socket)
            self.listener_socket = None
        return bool(self.conn_socket)

    def send(self, byte_array, flags=None):
        """
        Send data to the other end of the duct. NOTE: This send call is a thin wrapper over the underlying socket
        send call, and is thus _stream oriented_. A single send call may _not_ send all the data given to the method.
        It is your responsibility to add additional semantics and metadata to get "message-based" send/recv, and to
        check from the return value how many bytes were _actually_ sent.

        A NotConnectedException is raised if the duct hasn't been bound to the other end yet.

        :param byte_array: The bytes-like data to send to the other end.
        :type byte_array: bytearray | buffer | str | bytes
        :param flags: Optional flags to be set on the socket send call.
        :param flags: bytearray | buffer | str | bytes | None
        :return: The number of bytes sent.
        :rtype: int
        """
        if self.conn_socket is None:
            raise NotConnectedException("Must be connected to other end to send data!")
        if flags is None:
            return self.conn_socket.send(byte_array)
        else:
            return self.conn_socket.send(byte_array, flags)

    def recv(self, buff_size, flags=None):
        """
        Receive up to buff_size bytes from the remote host.

        A NotConnectedException is raised if the duct hasn't been bound to the other end yet.

        :param buff_size: The maximum number of bytes to read off the socket buffer from the remote host.
        :type buff_size: int
        :param flags: Optional flags to be set on the socket recv call.
        :param flags: bytearray | buffer | str | bytes | None
        :return: A bytes-like object that contains data from the remote host.
        :rtype: bytearray | buffer | str | bytes
        """
        if self.conn_socket is None:
            raise NotConnectedException("Must be connected to other end to receive data!")
        if flags is None:
            return self.conn_socket.recv(buff_size)
        else:
            return self.conn_socket.recv(buff_size, flags)
        
    def recv_into(self, buffer, recv_num_bytes=None, flags=None):
        """
        Receive up to buff_size bytes from the remote host. A buffer is given to write directly into to reduce the
        memory overhead of recv (one less copy needed).

        A NotConnectedException is raised if the duct hasn't been bound to the other end yet.

        :param buffer: An object that implements the buffer interface and can have data written directly into it.
        :type buffer: buffer
        :param recv_num_bytes: The maximum number of bytes to read off the socket buffer from the remote host.
            If this is not set, it is calculated from the length of the buffer passed in.
        :type recv_num_bytes: int | Nonre
        :param flags: Optional flags to be set on the socket recv call.
        :param flags: bytearray | buffer | str | bytes | None
        :return: The number of bytes read from the socket's recv buffer from the remote host.
        :rtype: int
        """
        if self.conn_socket is None:
            raise NotConnectedException("Must be connected to other end to receive data!")
        if recv_num_bytes is None and flags is None:
            return self.conn_socket.recv_into(buffer)
        elif flags is None and recv_num_bytes is not None:
            return self.conn_socket.recv_into(buffer, recv_num_bytes)
        elif flags is not None and recv_num_bytes is not None:
            return self.conn_socket.recv_into(buffer, recv_num_bytes, flags)            
        else:
            assert False

    def poll(self, timeout=60):
        """
        Poll to see if the socket has any data to read from the remote host.

        A NotConnectedException is raised if the duct hasn't been bound to the other end yet.

        A RemoteSocketClosed exception is raised if the local socket has faulted (we can't talk to the remote any more).

        :param timeout: Time to wait for data to show up. If 0, poll() does not block.
        :type timeout: float | int
        :return: True if this data to read, False otherwise.
        """
        if self.conn_socket is None:
            raise NotConnectedException("Must be connected to other end to poll for data!")
        has_recv_data, _, is_faulted = map(bool, select.select([self.conn_socket], [], [self.conn_socket], timeout))
        if is_faulted:
            self.server_connection_socket_destructor(self.conn_socket)
            raise RemoteSocketClosed("Remote socket is closed!")
        return has_recv_data

    def get_conn_file_descriptor(self):
        """
        Get the file descriptor of the underlying connection socket. This is useful for integrating into other event
        loops.

        A NotConnectedException is raised if the duct hasn't been bound to the other end yet.

        :return: The connection file descriptor.
        :rtype: int
        """
        if self.conn_socket is None:
            raise NotConnectedException("Must be connected to other end to have a file descriptor!")
        else:
            return self.conn_socket.fileno()

    def close(self):
        """
        Close the connection or listener sockets, if they're open.
        :return: None
        """
        if self.listener_socket is not None:
            self.server_listener_socket_destructor(self.listener_socket)
            self.listener_socket = None
        if self.conn_socket is not None:
            self.server_connection_socket_destructor(self.conn_socket)
            self.conn_socket = None

    def __del__(self):
        self.close()


class RawDuctChild(object):
    """
    The RawDuctChild is a thin wrapper over top of a "client" socket, and should be at the other end of a listening
    RawDuctParent. Once the first connection is seen, the listener socket is closed, and the "parent" and
    "child" act like an anonymous socket pair (with a few extra niceties, like a poll method).
    """

    DEFAULT_TIMEOUT = 30
    DEFAULT_CONNECT_RETRY_COUNT = 3
    DEFAULT_RETRY_DELAY = 3

    def __init__(self, connect_address, socket_constructor=unix_domain_socket_constructor,
                 socket_destructor=client_socket_destructor, timeout=DEFAULT_TIMEOUT):
        self.socket_constructor = socket_constructor
        self.socket_destructor = socket_destructor
        self.connect_address = connect_address
        self.socket = None
        self.socket_timeout = timeout

    def connect(self, connect_retry_count=DEFAULT_CONNECT_RETRY_COUNT, connect_retry_delay=DEFAULT_RETRY_DELAY):
        """
        Attempt to connect to another duct.

        AlreadyConnectedException is raised if the connection has already been established.

        :param connect_retry_count: The number of times to retry connecting if the connection is refused or if the
            file system entry for the Unix Domain socket does not yet exist on disk. Default: 3
        :type connect_retry_count: int
        :param connect_retry_delay: The amount of time to sleep between successive connect retries in the event of
            failure. Default: 3
        :type connect_retry_delay: int | float
        :return: None
        """
        if self.socket is not None:
            raise AlreadyConnectedException("Already connected to other end!")
        while True:
            try:
                self.socket = self.socket_constructor()
                self.socket.settimeout(self.socket_timeout)
                self.socket.connect(self.connect_address)
                return
            except socket.error as e:
                self.socket = None
                if e.errno == errno.ENOENT or e.errno == errno.ECONNREFUSED and connect_retry_count:
                    connect_retry_count -= 1
                    time.sleep(connect_retry_delay)
                else:
                    raise e

    def send(self, byte_array, flags=None):
        """
        Send data to the other end of the duct. NOTE: This send call is a thin wrapper over the underlying socket
        send call, and is thus _stream oriented_. A single send call may _not_ send all the data given to the method.
        It is your responsibility to add additional semantics and metadata to get "message-based" send/recv, and to
        check from the return value how many bytes were _actually_ sent.

        A NotConnectedException is raised if the duct hasn't been bound to the other end yet.

        :param byte_array: The bytes-like data to send to the other end.
        :type byte_array: bytearray | buffer | str | bytes
        :param flags: Optional flags to be set on the socket send call.
        :param flags: bytearray | buffer | str | bytes | None
        :return: The number of bytes sent.
        :rtype: int
        """
        if self.socket is None:
            raise NotConnectedException("Must be connected to other end to send data!")
        if flags is None:
            return self.socket.send(byte_array)
        else:
            return self.socket.send(byte_array, flags)

    def recv(self, buff_size, flags=None):
        """
        Receive up to buff_size bytes from the remote host.

        A NotConnectedException is raised if the duct hasn't been bound to the other end yet.

        :param buff_size: The maximum number of bytes to read off the socket buffer from the remote host.
        :type buff_size: int
        :param flags: Optional flags to be set on the socket recv call.
        :param flags: bytearray | buffer | str | bytes | None
        :return: A bytes-like object that contains data from the remote host.
        :rtype: bytearray | buffer | str | bytes
        """
        if self.socket is None:
            raise NotConnectedException("Must be connected to other end to receive data!")
        if flags is None:
            return self.socket.recv(buff_size)
        else:
            return self.socket.recv(buff_size, flags)

    def recv_into(self, buffer, recv_num_bytes=None, flags=None):
        """
        Receive up to buff_size bytes from the remote host. A buffer is given to write directly into to reduce the
        memory overhead of recv (one less copy needed).

        A NotConnectedException is raised if the duct hasn't been bound to the other end yet.

        :param buffer: An object that implements the buffer interface and can have data written directly into it.
        :type buffer: buffer
        :param recv_num_bytes: The maximum number of bytes to read off the socket buffer from the remote host.
            If this is not set, it is calculated from the length of the buffer passed in.
        :type recv_num_bytes: int | Nonre
        :param flags: Optional flags to be set on the socket recv call.
        :param flags: bytearray | buffer | str | bytes | None
        :return: The number of bytes read from the socket's recv buffer from the remote host.
        :rtype: int
        """
        if self.socket is None:
            raise NotConnectedException("Must be connected to other end to receive data!")
        if recv_num_bytes is None and flags is None:
            return self.socket.recv_into(buffer)
        elif flags is None and recv_num_bytes is not None:
            return self.socket.recv_into(buffer, recv_num_bytes)
        elif flags is not None and recv_num_bytes is not None:
            return self.socket.recv_into(buffer, recv_num_bytes, flags)
        else:
            assert False

    def poll(self, timeout=60):
        """
        Poll to see if the socket has any data to read from the remote host.

        A NotConnectedException is raised if the duct hasn't been bound to the other end yet.

        A RemoteSocketClosed exception is raised if the local socket has faulted (we can't talk to the remote any more).

        :param timeout: Time to wait for data to show up. If 0, poll() does not block.
        :type timeout: float | int
        :return: True if this data to read, False otherwise.
        """
        if self.socket is None:
            raise NotConnectedException("Must be connected to other end to poll for data!")
        has_recv_data, _, is_faulted = map(bool, select.select([self.socket], [], [self.socket], timeout))
        if is_faulted:
            self.socket_destructor(self.socket)
            raise RemoteSocketClosed("Remote socket is closed!")
        return has_recv_data

    def get_conn_file_descriptor(self):
        """
        Get the file descriptor of the underlying connection socket. This is useful for integrating into other event
        loops.

        A NotConnectedException is raised if the duct hasn't been bound to the other end yet.

        :return: The connection file descriptor.
        :rtype: int
        """
        if self.socket is None:
            raise NotConnectedException("Must be connected to other end to have a file descriptor!")
        else:
            return self.socket.fileno()

    def close(self):
        """
        Close the connection socket, if it's open.
        :return: None
        """
        if self.socket is not None:
            self.socket_destructor(self.socket)
            self.socket = None

    def __del__(self):
        self.close()



try:
    import anyjson as json
except ImportError:
    import json
import struct
import codecs
from binascii import hexlify
from tempfile import NamedTemporaryFile

from ductworks.base_duct import RawDuctParent, RawDuctChild, tcp_socket_constructor,\
    tcp_socket_listener_destructor


MAGIC_BYTE = b'\x54'


class MessageProtocolException(Exception):
    pass


def serializer_with_encoder_constructor(serialization_func, encoder_type='utf-8', encoder_error_mode='strict'):
    """
    Wrap a serialization function with string encoding. This is important for JSON, as it serializes objects into
    strings (potentially unicode), NOT bytestreams. An extra encoding step is needed to get to a bytestream.

    :param serialization_func: The base serialization function.
    :param encoder_type: The encoder type. Default: 'utf-8'
    :param encoder_error_mode: The encode error mode. Default: 'strict'.
    :return: The serializer function wrapped with specified encoder.
    """
    encoder = codecs.getencoder(encoder_type)

    def serialize(payload):
        serialized, _ = encoder(serialization_func(payload), encoder_error_mode)
        return serialized
    return serialize


def deserializer_with_encoder_constructor(deserialization_func, decoder_type='utf-8', decoder_error_mode='replace'):
    """
    Wrap a deserialization function with string encoding. This is important for JSON, as it expects to operate on
    strings (potentially unicode), NOT bytetsteams. A decoding steps is needed in between.

    :param deserialization_func: The base deserialization function.
    :param decoder_type: The decoder type. Default: 'utf-8'
    :param decoder_error_mode: The decode error mode. Default: 'replace'.
    :return: The deserializer function wrapped with specified decoder.
    """
    decoder = codecs.getdecoder(decoder_type)

    def deserialize(payload):
        decoded, _ = decoder(payload, decoder_error_mode)
        return deserialization_func(decoded)
    return deserialize


default_serializer = serializer_with_encoder_constructor(json.dumps)
default_deserializer = deserializer_with_encoder_constructor(json.loads)


class MessageDuctParent(object):
    def __init__(self, socket_duct, serialize=default_serializer, deserialize=default_deserializer, lock=None):
        self.socket_duct = socket_duct
        self.serialize = serialize
        self.deserialize = deserialize
        self.lock = lock

    @property
    def bind_address(self):
        return self.socket_duct.bind_address

    @property
    def listener_address(self):
        return self.socket_duct.listener_address

    def get_conn_file_descriptor(self):
        """
        Get the file descriptor for the connection socket in the socket duct.
        This is useful for integrating into other event loops.

        :return: The connection file descriptor.
        :rtype: int
        """
        return self.socket_duct.get_conn_file_descriptor()

    @classmethod
    def psuedo_anonymous_parent_duct(cls, bind_address=None, serialize=default_serializer,
                                     deserialize=default_deserializer, lock=None,
                                     timeout=RawDuctParent.DEFAULT_TIMEOUT):
        """
        Create a new psuedo-anonymous parent message duct with Unix Domain sockets.

        :param bind_address: The (filesystem) address to listen on, if any. If None a new random address will be
            chosen. Default: None.
        :type bind_address: basestring | None
        :param serialize: The serialization function for sending messages. Default: Encoded JSON.
        :param deserialize: The deserialization function for sending messages. Default: Encoded JSON.
        :param lock: A lock object to lock send/recv calls.
        :param timeout: The number of seconds to block a send/recv call waiting for completion.
        :type timeout: int | float
        :return: A new MessageDuctParent.
        :rtype: ductworks.MessageDuctParent
        """
        if bind_address is None:
            tmp = NamedTemporaryFile()
            bind_address = tmp.name
            tmp.close()
        return cls(
            RawDuctParent(bind_address=bind_address, timeout=timeout),
            serialize=serialize, deserialize=deserialize, lock=lock
        )

    @classmethod
    def psuedo_anonymous_tcp_parent_duct(cls, bind_address='localhost', bind_port=0, serialize=default_serializer,
                                         deserialize=default_deserializer, lock=None,
                                         timeout=RawDuctParent.DEFAULT_TIMEOUT):
        """
        Create a new psuedo-anonymous parent message duct with TCP sockets.

        :param bind_address: The interface address to listen on, if any. Default: 'localhost'.
        :type bind_address: basestring
        :param bind_port: The port to bind the interface to. If 0, pick the a random open port. Default: 0.
        :type bind_port: int
        :param serialize: The serialization function for sending messages. Default: Encoded JSON.
        :param deserialize: The deserialization function for sending messages. Default: Encoded JSON.
        :param lock: A lock object to lock send/recv calls.
        :param timeout: The number of seconds to block a send/recv call waiting for completion.
        :type timeout: int | float
        :return: A new MessageDuctParent.
        :rtype: ductworks.MessageDuctParent
        """

        return cls(
            RawDuctParent(
                bind_address=(bind_address, bind_port),
                server_listener_socket_constructor=tcp_socket_constructor,
                server_listener_socket_destructor=tcp_socket_listener_destructor,
                timeout=timeout
            ),
            serialize=serialize, deserialize=deserialize, lock=lock
        )

    def bind(self):
        """
        Bind the underlying socket duct.
        :return: None
        """
        return self.socket_duct.bind()

    def listen(self):
        """
        Listen on the underyling socket duct.
        :return: None
        """
        return self.socket_duct.listen()

    def poll(self, timeout=60):
        """
        Poll the underlying socket duct to check for new messages.
        :param timeout: The amount of time to wait for a new message, if none is present. Default: 60 seconds.
        :type timeout: int
        :return: True if a message is waiting, False otherwise.
        :rtype: bool
        """
        return self.socket_duct.poll(timeout)

    def send(self, payload):
        """
        Send a payload to the other end, if connected.

        :param payload: A serializable Python object to send to the other duct.
        :return: None
        """
        send_lock = self.lock
        try:
            if send_lock:
                send_lock.acquire()
            serialized_payload = self.serialize(payload)
            payload_len = len(serialized_payload)
            full_message = bytearray(struct.pack('!cL', MAGIC_BYTE, payload_len))
            full_message.extend(serialized_payload)
            full_message_view = memoryview(full_message)
            while full_message_view:
                bytes_sent = self.socket_duct.send(full_message_view)
                full_message_view = full_message_view[bytes_sent:]
        finally:
            if send_lock:
                send_lock.release()

    def recv(self):
        """
        Receive a payload from the other end, if connected and data is present.

        :return: A deserialized Python object from the other end of the duct.
        """
        recv_lock = self.lock
        try:
            if recv_lock:
                recv_lock.acquire()
            leading_byte = self.socket_duct.recv(1)
            if leading_byte != MAGIC_BYTE:
                raise MessageProtocolException("Invalid magic byte at message envelope head! Expected: {}, got: {}"
                                               "".format(hexlify(MAGIC_BYTE), hex(ord(leading_byte))))
            raw_payload_len = bytearray(struct.calcsize('!L'))
            raw_payload_len_view = memoryview(raw_payload_len)
            while raw_payload_len_view:
                num_bytes_recieved = self.socket_duct.recv_into(raw_payload_len_view)
                raw_payload_len_view = raw_payload_len_view[num_bytes_recieved:]
            incoming_payload_len, = struct.unpack('!L', bytes(raw_payload_len))
            serialized_payload = bytearray(incoming_payload_len)
            serialized_payload_view = memoryview(serialized_payload)
            while serialized_payload_view:
                num_bytes_recieved = self.socket_duct.recv_into(serialized_payload_view)
                serialized_payload_view = serialized_payload_view[num_bytes_recieved:]
            return self.deserialize(serialized_payload)
        finally:
            if recv_lock:
                recv_lock.release()

    def close(self):
        """
        Close the underlying socket duct.
        :return: None
        """
        self.socket_duct.close()

    def __del__(self):
        self.close()


class MessageDuctChild(object):
    def __init__(self, socket_duct, serialize=default_serializer, deserialize=default_deserializer, lock=None):
        self.socket_duct = socket_duct
        self.serialize = serialize
        self.deserialize = deserialize
        self.lock = lock

    @property
    def connect_address(self):
        return self.socket_duct.connect_address

    def get_conn_file_descriptor(self):
        """
        Get the file descriptor for the connection socket in the socket duct.
        This is useful for integrating into other event loops.

        :return: The connection file descriptor.
        :rtype: int
        """
        return self.socket_duct.get_conn_file_descriptor()

    @classmethod
    def psuedo_anonymous_child_duct(cls, connect_address, serialize=default_serializer,
                                    deserialize=default_deserializer, lock=None,
                                    timeout=RawDuctChild.DEFAULT_TIMEOUT):
        """
        Create a new psuedo-anonymous child message duct with Unix Domain sockets. The connect_address parameter
        should be sourced from the parent duct by getting its listener_address property.

        :param connect_address: The filesystem address to connect to. Get this from parent_duct.listener_address.
        :type connect_address: basestring
        :param serialize: The serialization function for sending messages. Default: Encoded JSON.
        :param deserialize: The deserialization function for sending messages. Default: Encoded JSON.
        :param lock: A lock object to lock send/recv calls.
        :param timeout: The number of seconds to block a send/recv call waiting for completion.
        :type timeout: int | float
        :return: A new MessageParentDuct.
        :rtype: ductworks.MessageDuctChild
        """
        return cls(
            RawDuctChild(
                connect_address, timeout=timeout
            ),
            serialize=serialize, deserialize=deserialize, lock=lock
        )

    @classmethod
    def psuedo_anonymous_tcp_child_duct(cls, connect_address, connect_port, serialize=default_serializer,
                                        deserialize=default_deserializer, lock=None,
                                        timeout=RawDuctChild.DEFAULT_TIMEOUT):
        """
        Create a new psuedo-anonymous child message duct with Unix Domain sockets. The connect_address and
        connect_port parameters should be sourced from the parent duct by getting its listener_address property.

        :param connect_address: The interface address to connect to. Get this from parent_duct.listener_address[0].
        :type connect_address: basestring
        :param connect_port: The TCP port to connect to. Get this from parent_duct.listener_address[1].
        :type connect_port: int
        :param serialize: The serialization function for sending messages. Default: Encoded JSON.
        :param deserialize: The deserialization function for sending messages. Default: Encoded JSON.
        :param lock: A lock object to lock send/recv calls.
        :param timeout: The number of seconds to block a send/recv call waiting for completion.
        :type timeout: int | float
        :return: A new MessageParentDuct.
        :rtype: ductworks.MessageDuctChild
        """

        return cls(
            RawDuctChild(
                connect_address=(connect_address, connect_port),
                socket_constructor=tcp_socket_constructor,
                timeout=timeout
            ),
            serialize=serialize, deserialize=deserialize, lock=lock
        )

    def connect(self):
        """
        Call connect() on the underlying socket duct.
        :return: None
        """
        return self.socket_duct.connect()

    def poll(self, timeout=60):
        """
        Poll the underlying socket duct to check for new messages.
        :param timeout: The amount of time to wait for a new message, if none is present. Default: 60 seconds.
        :type timeout: int
        :return: True if a message is waiting, False otherwise.
        :rtype: bool
        """
        return self.socket_duct.poll(timeout)

    def send(self, payload):
        """
        Send a payload to the other end, if connected.

        :param payload: A serializable Python object to send to the other duct.
        :return: None
        """
        send_lock = self.lock
        try:
            if send_lock:
                send_lock.acquire()
            serialized_payload = self.serialize(payload)
            payload_len = len(serialized_payload)
            full_message = bytearray(struct.pack('!cL', MAGIC_BYTE, payload_len))
            full_message.extend(serialized_payload)
            full_message_view = memoryview(full_message)
            while full_message_view:
                bytes_sent = self.socket_duct.send(full_message_view)
                full_message_view = full_message_view[bytes_sent:]
        finally:
            if send_lock:
                send_lock.release()

    def recv(self):
        """
        Receive a payload from the other end, if connected and data is present.

        :return: A deserialized Python object from the other end of the duct.
        """
        recv_lock = self.lock
        try:
            if recv_lock:
                recv_lock.acquire()
            leading_byte = self.socket_duct.recv(1)
            if leading_byte != MAGIC_BYTE:
                raise MessageProtocolException("Invalid magic byte at message envelope head! Expected: {}, got: {}"
                                               "".format(hexlify(MAGIC_BYTE), hex(ord(leading_byte))))
            raw_payload_len = bytearray(struct.calcsize('!L'))
            raw_payload_len_view = memoryview(raw_payload_len)
            while raw_payload_len_view:
                num_bytes_recieved = self.socket_duct.recv_into(raw_payload_len_view)
                raw_payload_len_view = raw_payload_len_view[num_bytes_recieved:]
            incoming_payload_len, = struct.unpack('!L', bytes(raw_payload_len))
            serialized_payload = bytearray(incoming_payload_len)
            serialized_payload_view = memoryview(serialized_payload)
            while serialized_payload_view:
                num_bytes_recieved = self.socket_duct.recv_into(serialized_payload_view)
                serialized_payload_view = serialized_payload_view[num_bytes_recieved:]
            return self.deserialize(serialized_payload)
        finally:
            if recv_lock:
                recv_lock.release()

    def close(self):
        """
        Close the underlying socket duct.
        :return: None
        """
        self.socket_duct.close()

    def __del__(self):
        self.close()


def create_psuedo_anonymous_duct_pair(serialize=default_serializer, deserialize=default_deserializer,
                                      parent_lock=None, child_lock=None):
    """
    Create an already connected pair of anonymous ducts. This is very similar to how multiprocess.Pipe(True) functions.

    :param serialize: The serializer function for the pair. Defaults to encoded JSON.
    :param deserialize: The deserializer funtion for the pair. Defaults to encoded JSON.
    :param parent_lock: An optional lock object to give to the "parent" duct.
    :param child_lock: An optional lock object to give to the "child" duct.
    :return: A parent/child pair of ducts.
    :rtype: (ductworks.MessageDuctParent, ductworks.MesssageDuctChild)
    """

    parent = MessageDuctParent.psuedo_anonymous_parent_duct(
        serialize=serialize, deserialize=deserialize, lock=parent_lock
    )
    parent.bind()
    listener_address = parent.listener_address
    child = MessageDuctChild.psuedo_anonymous_child_duct(
        listener_address, serialize=serialize, deserialize=deserialize, lock=child_lock
    )
    child.connect()
    parent.listen()
    return parent, child



from setuptools import setup
from ductworks import __version__, __author__, __author_email__

setup(
    name='ductworks',
    author=__author__,
    author_email=__author_email__,
    url="https://bitbucket.org/rpcope1/ductworks",
    version=__version__,
    description="A simple IPC library for less simple situations.",
    long_description="""
    Ductworks is an IPC library for Python, similar to the Pipe object from the multiprocessing library. Like Pipe,
    ductworks exposes a MessageDuct API with very simple send and receive semantics for end users (no worrying about
    streams, each send call corresponds to one recv call), while easily allowing for more flexible usage like having
    a communication channel between a parent python process and a python process in a child subprocess (separated by
    both fork(2) and exec(2)). As well ductworks allows users to freely choose their serialization library (and uses
    JSON by default) and uses regular BSD sockets; this makes connecting up programs in different languages (or even on
    different systems) relatively straightforward. Ductworks aims to make it simple to get socket pairs, when inheriting
    a anonymous socket pair is difficult or outright infeasible, and to make IPC easier and less stressful.
    """,
    packages=['ductworks'],
    install_requires=[],
    tests_require=['assertpy'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Networking',
        'License :: OSI Approved :: BSD License'
    ]
)
